package main

import (
	"flag"
	"fmt"
	"os"
	"runtime"
)

var version = Version{major: 2, minor: 1, patch: 0}

type Version struct {
	major int
	minor int
	patch int
}

func (v Version) String() string {
	return fmt.Sprintf("%d.%d.%d", v.major, v.minor, v.patch)
}

type Config struct {
	name        string
	callname    string
	version     Version
	threads     int
	workers     int
	cpuprofile  string
	programfile string
	showversion bool
	query       string
	benchmark   bool
	solutions   int
	output      bool
	//interactive    bool
	sorted           bool
	showcoinduct     bool
	guardcheck       bool
	makequery        bool
	debug            bool
	showgroundroot   bool
	showrootsubst    bool
	maxsubstitutions int
	occurscheck      bool
}

var config Config = Config{
	name:        "goalp",
	callname:    os.Args[0],
	version:     version,
	threads:     runtime.NumCPU(),
	workers:     runtime.NumCPU(),
	cpuprofile:  "",
	programfile: "",
	showversion: false,
	query:       "",
	benchmark:   false,
	solutions:   0,
	output:      true,
	//interactive:    false,
	sorted:           false,
	showcoinduct:     false,
	guardcheck:       true,
	makequery:        true,
	debug:            false,
	showgroundroot:   false,
	showrootsubst:    false,
	maxsubstitutions: 0,
	occurscheck:      true,
}

func init() {
	flag.BoolVar(&config.showversion, "v", config.showversion, "show version")
	flag.BoolVar(&config.debug, "debug", config.showversion, "show debug output")
	flag.BoolVar(&config.showgroundroot, "showground", config.showgroundroot, "show trees with ground root")
	flag.BoolVar(&config.showrootsubst, "showrootsubst", config.showrootsubst, "show trees which had a substitution applied to the root")
	flag.BoolVar(&config.benchmark, "benchmark", config.benchmark, "print elapsed time and stops when requested number of solutions reached")
	flag.BoolVar(&config.output, "output", config.output, "print solutions")
	flag.BoolVar(&config.guardcheck, "guardcheck", config.guardcheck, "check guardedness of input program")
	flag.IntVar(&config.solutions, "solutions", config.solutions, "maximum number of solutions (0 means infinitely many)")
	flag.IntVar(&config.maxsubstitutions, "maxsubstitutions", config.maxsubstitutions, "maximum number of substitutions (0 means infinitely many)")
	flag.StringVar(&config.cpuprofile, "cpuprofile", config.cpuprofile, "write cpu profile to file")
	flag.StringVar(&config.query, "query", config.query, "query string")
	flag.StringVar(&config.programfile, "program", config.programfile, "read rules from file")
	flag.IntVar(&config.threads, "threads", config.threads, "number of threads to use")
	flag.IntVar(&config.workers, "workers", config.workers, "number of workers to use")
	//flag.BoolVar(&config.interactive, "interactive", config.interactive, "prompts at each expansion of a node corresponding to a coinductive predicate")
	flag.BoolVar(&config.sorted, "sorted", config.sorted, "sort solutions by number of expansions")
	flag.BoolVar(&config.occurscheck, "occurscheck", config.occurscheck, "activate occurs check in unifications")
}

func parseFlags() {
	flag.Usage = func() {
		fmt.Fprintf(os.Stderr, "usage: %s [options] PROGRAMFILE [query]\n", config.callname)
		flag.PrintDefaults()
		os.Exit(2)
	}

	flag.Parse()

	if config.showversion {
		fmt.Printf("%s version %s\n", config.name, config.version)
		os.Exit(0)
	}

	switch flag.NArg() {
	case 0:
		flag.Usage()
	case 1:
		config.makequery = false
	default:
		config.query = flag.Arg(1)
	}

	config.programfile = flag.Arg(0)

	//if config.interactive {
	//	config.workers = 1
	//}

}
