package main

import (
	"fmt"
	"os"
	"sort"
	"strconv"
	"sync/atomic"
)

type CoTree struct {
	Unifications int
	Tree         *OrNode
	Query        Query
}

func (cotree CoTree) printSolution() {
	if config.output {
		fmt.Print("solution: ", cotree.Unifications, " ", cotree.Query.prettyString(gpidmap))
	}
}

func NewStartCoTree(query Query, tmap NodeTemplateMap) CoTree {
	andnode := NewAndNode(make(OrNodes, len(query)))
	for i, queryterm := range query {
		open := tmap[queryterm.Id()]
		openFinal := make(NodeTemplates, 0, len(open))
		for _, node := range open {
			if node.Head.Unifiable(queryterm) {
				openFinal = append(openFinal, node)
			}
		}
		andnode.And[i] = NewOrNode(queryterm, nil, openFinal).Unroll()
	}
	return CoTree{0, NewOrNode(NewPredicate(queryPredId, nil), AndNodes{andnode}, nil), query}
}

// make substitutions be a id / term not vidmap
func (cotree CoTree) expansions(open *OrNode) []CoTree {
	if open == nil {
		return nil
	}
	substslist := open.SubstitutionsList()
	newcotrees := make([]CoTree, 0, len(substslist)+1)
	for _, substs := range substslist {
		newcotrees = append(newcotrees, cotree.CopyExpand(substs.Expand()))
	}
	//TODO: if there is only one open leave we do not need the additional copy
	open.Open = nil
	return append(newcotrees, cotree)
}

func (cotree CoTree) CopyExpand(substs TermMap) CoTree {
	substsCopy := substs.Copy()
	newcotree := CoTree{cotree.Unifications + 1, cotree.Tree.CopyUnroll(substsCopy), cotree.Query.copyMap(substsCopy)}
	return newcotree
}

type NodeTemplateMap map[Id](NodeTemplates)

type NodeTemplates []*NodeTemplate

func (nts NodeTemplateMap) Ids() []Id {
	ids := make([]Id, 0, len(nts))
	for key := range nts {
		ids = append(ids, key)
	}
	sort.Sort(IdSlice(ids))
	return ids
}

func (nts NodeTemplates) String() string {
	if len(nts) == 0 {
		return "[]"
	}
	s := fmt.Sprintf("[%s", nts[0].Head)
	for _, nt := range nts[1:] {
		s = fmt.Sprintf("%s,%s", s, nt.Head)
	}
	return s + "]"
}

func (nts NodeTemplates) Copy() NodeTemplates {
	ntsNew := make(NodeTemplates, len(nts))
	copy(ntsNew, nts)
	return ntsNew
}

type NodeTemplate struct {
	Head *Term
	Body *AndNode
}

func NewNodeTemplate(head *Term, body *AndNode) *NodeTemplate {
	return &NodeTemplate{head, body}
}

func (nt NodeTemplate) String() string {
	return fmt.Sprintf("%s->%s", nt.Head, nt.Body)
}

func (nt *NodeTemplate) makeAndNode(head *Term) *AndNode {
	tmap := make(TermMap)
	nt.Head.Copy_(tmap).Match(head)
	andnodes := nt.Body.And //TODO: Not sure about whether to make a copy here or reference or nothing for speedup
	ornodes := make(OrNodes, len(andnodes))
	for i, ornode := range andnodes {
		ornodes[i] = NewOrNode(ornode.Head.Copy_(tmap), nil, ornode.Open.Copy())
	}
	return NewAndNode(ornodes)
}

type OrNodes []*OrNode

type OrNode struct {
	Head *Term
	Or   AndNodes
	Open NodeTemplates
}

func NewOrNode(head *Term, andnodes AndNodes, open NodeTemplates) *OrNode {
	return &OrNode{head, andnodes, open}
}

func (n OrNode) String() string {
	return fmt.Sprintf("OR{%s|$or%s|$open%s}", n.Head, n.Or, n.Open)
}

func (n OrNode) prettyString(pidmap PidMap) string {
	s := n.Head.prettyString(pidmap)
	if len(n.Open) != 0 {
		s += "?"
	}
	if len(n.Or) == 0 {
		return s
	}
	if len(n.Or) == 1 {
		if len(n.Or[0].And) == 0 {
			return s
		}
		return fmt.Sprintf("%s->%s", s, n.Or[0].prettyString(pidmap))
	}
	s = fmt.Sprintf("%s->{%s", s, n.Or[0].prettyString(pidmap))
	for _, a := range n.Or[1:] {
		s = fmt.Sprintf("%s,%s", s, a.prettyString(pidmap))
	}
	return s + "}"
}

func (n *OrNode) isSolved() bool {
	for _, andnode := range n.Or {
		if andnode.isSolved() {
			return true
		}
	}
	return false
}

func (n *OrNode) canBeSolved() bool {
	if len(n.Open) > 0 {
		return true
	}
	for _, andnode := range n.Or {
		if andnode.canBeSolved() {
			return true
		}
	}
	return false
}

func (n *OrNode) CopyUnroll(subst TermMap) *OrNode {
	if len(n.Or) == 1 && len(n.Or[0].And) == 1 && len(n.Open) == 0 {
		return n.Or[0].And[0].CopyUnroll(subst)
	}
	headNew := n.Head.Copy_(subst)
	orNew := make(AndNodes, len(n.Or))
	openNew := make(NodeTemplates, 0, len(n.Open))
	for i := range n.Or {
		orNew[i] = &AndNode{make(OrNodes, len(n.Or[i].And))}
		for k := range n.Or[i].And {
			orNew[i].And[k] = n.Or[i].And[k].CopyUnroll(subst)
		}
	}
	for _, open := range n.Open {
		//TODO: optimize, check unifiable and subsumes in one go
		//TODO: get rid of this Head.Copy by having threadsafe tmap?
		ntsHeadCopy := open.Head.Copy()
		if ntsHeadCopy.Unifiable(headNew) {
			if ntsHeadCopy.Subsumes(headNew) {
				orNew = append(orNew, open.makeAndNode(headNew).Unroll())
			} else {
				openNew = append(openNew, open)
			}
		}
	}
	return NewOrNode(headNew, orNew, openNew)
}

func (n *OrNode) UnrollOpen() {
	//TODO: optimize away multi append by shifting nodes in open only once
	for i := 0; i < len(n.Open); {
		ntsHeadCopy := n.Open[i].Head.Copy()
		if ntsHeadCopy.Unifiable(n.Head) {
			if ntsHeadCopy.Subsumes(n.Head) {
				n.Or = append(n.Or, n.Open[i].makeAndNode(n.Head))
				n.Open = append(n.Open[:i], n.Open[i+1:]...)
			} else {
				i++
			}
		} else {
			n.Open = append(n.Open[:i], n.Open[i+1:]...)
		}
	}
}

func (n *OrNode) Unroll() *OrNode {
	n.UnrollOpen()
	n.Or.Unroll()
	return n
}

func (n *OrNode) findOpen(level int) (rlevel int, ropennode *OrNode, rcoinductive bool) {
	if len(n.Open) > 0 {
		_, rcoinductive = gcoindpidmap[n.Head.Id()]
		return level, n, rcoinductive
	}
	for _, andnode := range n.Or {
		for _, ornode := range andnode.And {
			nlevel, opennode, coinductive := ornode.findOpen(level + 1)
			if (opennode != nil) && (nlevel < rlevel || ropennode == nil || (!coinductive && rcoinductive)) {
				ropennode = opennode
				rlevel = nlevel
				rcoinductive = coinductive
			}
		}
	}
	return rlevel, ropennode, rcoinductive
}

func (n *OrNode) SubstitutionsList() []TermMap {
	substslist := make([]TermMap, len(n.Open))
	for i := range n.Open {
		substslist[i] = TermSubstitutions(n.Head, n.Open[i].Head.Copy())
	}
	return substslist
}

func (n *OrNode) Copy_(tmap TermMap) *OrNode {
	return NewOrNode(n.Head.Copy_(tmap), n.Or.Copy_(tmap), n.Open.Copy())
}

type AndNodes []*AndNode

func (andnodes AndNodes) Copy_(tmap TermMap) AndNodes {
	andnodesCopy := make(AndNodes, len(andnodes))
	for i := range andnodes {
		andnodesCopy[i] = andnodes[i].Copy_(tmap)
	}
	return andnodesCopy
}

func (andnodes AndNodes) Unroll() {
	for _, andnode := range andnodes {
		andnode.Unroll()
	}
}

type AndNode struct {
	And OrNodes
}

func (andnode *AndNode) Unroll() *AndNode {
	for _, ornode := range andnode.And {
		ornode.Unroll()
	}
	return andnode
}

func NewAndNode(ornodes OrNodes) *AndNode {
	return &AndNode{And: ornodes}
}

func (n AndNode) String() string {
	if len(n.And) == 0 {
		return "AND[]"
	}
	s := fmt.Sprintf("[%s", n.And[0])
	for _, a := range n.And[1:] {
		s = fmt.Sprintf("%s,%s", s, a)
	}
	return s + "]"
}

func (n AndNode) prettyString(pidmap PidMap) string {
	if len(n.And) == 0 {
		return "!"
	}
	s := fmt.Sprintf("[%s", n.And[0].prettyString(pidmap))
	for _, a := range n.And[1:] {
		s = fmt.Sprintf("%s,%s", s, a.prettyString(pidmap))
	}
	return s + "]"
}

func (n *AndNode) isSolved() bool {
	for _, ornode := range n.And {
		if !ornode.isSolved() {
			return false
		}
	}
	return true
}

func (n *AndNode) canBeSolved() bool {
	for _, ornode := range n.And {
		if !ornode.canBeSolved() {
			return false
		}
	}
	return true
}

func (n *AndNode) Copy_(tmap TermMap) *AndNode {
	ornodesCopy := make(OrNodes, len(n.And))
	for i := range n.And {
		ornodesCopy[i] = n.And[i].Copy_(tmap)
	}
	return NewAndNode(ornodesCopy)
}

func makeTemplates(input string, pmap PnameMap) NodeTemplateMap {
	tmap := make(NodeTemplateMap)
	prules, err := ParseRules(input)
	if err != nil {
		panic(err)
	}
	for _, prule := range prules {
		vmap := make(VnameMap)
		if prule.Head == nil {
			if len(prule.Body) == 1 && prule.Body[0].Name == "coinductive" &&
				len(prule.Body[0].Args) == 1 && prule.Body[0].Args[0].Name == "/" && len(prule.Body[0].Args[0].Args) == 2 &&
				len(prule.Body[0].Args[0].Args[0].Args) == 0 && len(prule.Body[0].Args[0].Args[1].Args) == 0 {
				name := prule.Body[0].Args[0].Args[0].Name

				arity, err := strconv.Atoi(prule.Body[0].Args[0].Args[1].Name)
				if err != nil {
					fmt.Println("unrecognized directive: :-", makeTermM(prule.Body[0], pmap, vmap).prettyString(pmap.PidMap()))
					os.Exit(1)
				}
				key := ptermKey{name, arity}
				id, ok := pmap[key]
				if !ok {
					id = Id(atomic.AddInt64((*int64)(&genid), 1))
					pmap[key] = id
				}
				gcoindpidmap[id] = name
			} else {
				fmt.Println("unrecognized directive: :-", makeTermM(prule.Body[0], pmap, vmap).prettyString(pmap.PidMap()))
				os.Exit(1)
			}
		} else {
			head := makeTermM(prule.Head, pmap, vmap)
			body := NewAndNode(make(OrNodes, len(prule.Body)))
			for k, term := range prule.Body {
				termNew := makeTermM(term, pmap, vmap)
				body.And[k] = NewOrNode(termNew, nil, nil)
			}
			id := head.Id()
			nodetemplates, ok := tmap[id]
			if !ok {
				nodetemplates = nil
			}
			tmap[id] = append(nodetemplates, NewNodeTemplate(head, body))
		}
	}
	for _, nodetemplates := range tmap {
		for _, nodetemplate := range nodetemplates {
			for _, ornode := range nodetemplate.Body.And {
				// Open of Nodetemplates in tmap connect cyclic back to other Nodetemplates in tmap
				open := tmap[ornode.Head.Id()]
				openFinal := make(NodeTemplates, 0, len(open))
				for _, node := range open {
					if node.Head.UnifiableNoCheck(ornode.Head) {
						openFinal = append(openFinal, node)
					}
				}
				ornode.Open = openFinal
			}
		}
	}
	return tmap
}
