package main

import (
	"bufio"
	"container/heap"
	"fmt"
	"io/ioutil"
	"log"
	"os"
	"runtime"
	"runtime/pprof"
	"strconv"
	"strings"
	"sync"
	"time"
)

const MaxInt = int(^uint(0) >> 1)

var reader *bufio.Reader = bufio.NewReader(os.Stdin)

var parralelexpandworker int64

var gpidmap PidMap

var gcoindpidmap PidMap = make(PidMap)

func readProgram(filename string) string {
	bytes, err := ioutil.ReadFile(filename)
	if err != nil {
		log.Fatalf("cant find/open file %s", filename)
	}
	return string(bytes)
}

type workerdone struct {
	id      int
	cotrees []CoTree
}

type workerwait struct {
	id int
	in chan CoTree
}

func workParallel(query CoTree, wg *sync.WaitGroup) {
	exit := false
	searchqueue := &PriorityQueue{}
	solutionqueue := &PriorityQueue{}
	workerqueue := make(chan workerwait, config.workers)
	workerqueuewithstate := workerqueue
	workeroutqueue := make(chan workerdone, config.workers)
	solutionsinqueue := make(chan CoTree, config.workers)
	workerpriorities := make([]int, config.workers)
	var solutionminpriority int
	heap.Push(searchqueue, query)
	for i := 0; i < config.workers; i++ {
		workqueue := make(chan CoTree)
		wg.Add(1)
		go Worker(i, workerqueue, workqueue, workeroutqueue, solutionsinqueue, wg)
		workerqueue <- workerwait{i, workqueue}
	}
	var solutions int
	for (config.solutions == 0 || solutions < config.solutions) && !exit {
		select {
		case solution := <-solutionsinqueue:
			{
				solutions++
				if config.maxsubstitutions == 0 || config.maxsubstitutions >= solution.Unifications {
					if !config.sorted && !config.benchmark {
						solution.printSolution()
					} else {
						heap.Push(solutionqueue, solution)
					}
					if searchqueue.Len() > 0 {
						workerqueuewithstate = workerqueue
					}
				}
				if !config.benchmark && (solutions == config.solutions || config.solutions == 0) {
					solutionqueue.PrintSolutions(MaxInt)
					solutions = solutions - config.solutions
					fmt.Printf("number of requested solutions reached; continue [y/n/number]? ")
					line, _ := reader.ReadString('\n')
					line = strings.TrimSpace(line)
					if line == "n" || line == "N" {
						exit = true
					}
					newmaxsolutions, err := strconv.Atoi(line)
					if err == nil {
						config.solutions = solutions + newmaxsolutions
					}
				}
				if (len(workerqueue) == config.workers) && (len(workeroutqueue) == 0) && (searchqueue.Len() == 0) {
					exit = true
				}
			}
		case workerout := <-workeroutqueue:
			{
				workerpriorities[workerout.id] = MaxInt
				solutionminpriority = minIntSlice(workerpriorities)
				if config.maxsubstitutions == 0 || (len(workerout.cotrees) > 0 && config.maxsubstitutions >= workerout.cotrees[0].Unifications) {
					for _, newcotree := range workerout.cotrees {
						heap.Push(searchqueue, newcotree)
						if config.debug {
							fmt.Println("added to search queue: ", newcotree.Unifications, " ", newcotree.Query.prettyString(gpidmap))
							fmt.Println(newcotree.Tree.prettyString(gpidmap))
							fmt.Println("search queue length: ", searchqueue.Len())
						}
					}
				}
				if searchqueue.Len() > 0 {
					workerqueuewithstate = workerqueue
				}
				if (len(workerqueue) == config.workers) && (len(workeroutqueue) == 0) && (searchqueue.Len() == 0) {
					exit = true
				}
			}
		case workerin := <-workerqueuewithstate:
			{
				smin := heap.Pop(searchqueue).(CoTree)
				workerpriorities[workerin.id] = smin.Unifications
				solutionminpriority = minIntSlice(workerpriorities)
				workerin.in <- smin
				if searchqueue.Len() == 0 {
					workerqueuewithstate = nil
				}
				if config.sorted && !config.benchmark {
					solutionqueue.PrintSolutions(solutionminpriority)
				}
			}
		}
		runtime.Gosched()
	}
	if !config.benchmark {
		solutionqueue.PrintSolutions(MaxInt)
	}
	for {
		select {
		case workerin := <-workerqueue:
			{
				close(workerin.in)
			}
		default:
			return
		}
	}
}

func Worker(id int, workerqueue chan workerwait, in chan CoTree, out chan workerdone, solutions chan CoTree, wg *sync.WaitGroup) {
	var ok bool
	var loopcounter int
	var cotree CoTree
	var opennode *OrNode
	cotree, ok = <-in
	if !ok {
		wg.Done()
		return
	}
	for {
		if cotree.Tree.isSolved() {
			solutions <- cotree
		} else {
			var coinductive bool
			_, opennode, coinductive = cotree.Tree.findOpen(0)
			if opennode != nil && coinductive && !config.benchmark {
				fmt.Printf("expand coinductive node %s in %s with each substitution from list %s [y/n]?\n", opennode.Head.prettyString(gpidmap), cotree.Tree.prettyString(gpidmap), opennode.SubstitutionsList())
				line, _ := reader.ReadString('\n')
				line = strings.TrimSpace(line)
				if line == "n" || line == "N" {
					opennode.Open = make(NodeTemplates, 0)
				}
			}
			cotrees := cotree.expansions(opennode)
			if config.showgroundroot {
				for _, expcotree := range cotrees {
					ground := true
					subsumes := true
					for i, term := range expcotree.Query {
						if !term.isGround() {
							ground = false
							break
						}
						if !term.Subsumes(cotree.Query[i]) {
							subsumes = false
						}
					}
					if ground && !subsumes {
						fmt.Printf("ground tree root: %s\n", expcotree.Query.prettyString(gpidmap))
					}
				}
			}
			if config.showrootsubst {
				for _, expcotree := range cotrees {
					subsumes := true
					for i, term := range expcotree.Query {
						if !term.Subsumes(cotree.Query[i]) {
							subsumes = false
						}
					}
					if !subsumes {
						fmt.Printf("root with new substitution: %s\n", expcotree.Query.prettyString(gpidmap))
					}
				}
			}
			newCotrees := make([]CoTree, 0, len(cotrees))
			for _, cotree := range cotrees {
				if cotree.Tree.canBeSolved() {
					newCotrees = append(newCotrees, cotree)
				}
			}
			cotrees = newCotrees
			if len(cotrees) == 1 {
				if loopcounter < 10 {
					cotree = cotrees[0]
					loopcounter++
					continue
				} else {
					loopcounter = 0
				}
			}
			out <- workerdone{id, cotrees}
		}
		workerqueue <- workerwait{id, in}
		cotree, ok = <-in
		if !ok {
			wg.Done()
			break
		}
	}
}

func main() {
	var m runtime.MemStats
	runtime.ReadMemStats(&m)
	parseFlags()
	runtime.GOMAXPROCS(config.threads)
	if config.cpuprofile != "" {
		f, err := os.Create(config.cpuprofile)
		if err != nil {
			log.Fatal(err)
		}
		pprof.StartCPUProfile(f)
		defer pprof.StopCPUProfile()
	}
	names := make(PnameMap)
	tmap := makeTemplates(readProgram(config.programfile), names)
	gpidmap = names.PidMap()
	if config.guardcheck {
		var start time.Time
		if config.benchmark {
			log.Println("starting guardednesschecks")
			start = time.Now()
		}
		guardCheckProgram(tmap)
		if config.benchmark {
			log.Printf("guardednesschecks completed in %s", time.Since(start))
		}
	}

	if config.makequery {
		var start time.Time
		var wg = new(sync.WaitGroup)
		if config.benchmark {
			log.Println("start with", config.workers, "workers")
			start = time.Now()
		}
		query := makeQuery(config.query, names)
		gpidmap = names.PidMap()
		goal := NewStartCoTree(query, tmap)
		workParallel(goal, wg)
		if config.benchmark {
			timeTrack(start)
		}
	}
}
