package main

import (
	"fmt"
	"os"
)

type path []int

func (p path) equals(p2 path) bool {
	if len(p) != len(p2) {
		return false
	}
	for i := range p {
		if p[i] != p2[i] {
			return false
		}
	}
	return true
}

type gtrail struct {
	id           Id
	clause       int
	term         *Term
	projections  []context
	guardcontext []context
}

type context struct {
	clause  int
	measure *Term
	node    path
}

func (p path) String() string {
	var s string
	if len(p) == 0 {
		return "[]"
	}
	for _, u := range p {
		s = fmt.Sprintf("%s ,%d", s, u)
	}
	return "[" + s[2:] + "]"
}

func (c context) String() string {
	return fmt.Sprintf("{%d | %s | %s}", c.clause, c.measure.prettyString(gpidmap), c.node)
}

func (c context) matches(c2 context) bool {
	return c.measure.Subsumes(c2.measure)
}

func (c context) equals(c2 context) bool {
	return c.clause == c2.clause && c.node.equals(c2.node) && c.measure.Equals(c2.measure)
}

func trailPath(gtrail []gtrail) string {
	var s string
	if len(gtrail) == 0 {
		return ""
	}
	for _, g := range gtrail {
		s = fmt.Sprintf("%s, (%s:%d)", s, gpidmap[g.id], g.clause+1)
	}
	return s[2:]
}

func equalContext(g1, g2 []context) bool {
	if len(g1) != len(g2) {
		return false
	}
	gcover := make([]context, len(g2))
	copy(gcover, g2)
	for _, c1 := range g1 {
		for i, c2 := range gcover {
			if c1.equals(c2) {
				gcover = append(gcover[:i], gcover[i+1:]...)
				break
			}
		}
	}
	if len(gcover) == 0 {
		return true
	}
	return false
}

func guardCheckProgram(tmap NodeTemplateMap) {
	for _, id := range tmap.Ids() {
		for j, nodetemplate := range tmap[id] {
			if config.debug {
				fmt.Printf("checking clause %d with predicate '%s': %s\n", j+1, gpidmap[id], nodetemplate.Head.prettyString(gpidmap))
			}
			ginit := make([]gtrail, 0)

			if t, trail := isGuarded(tmap, nodetemplate.Head.Copy(), ginit); t != nil {
				fmt.Printf("non guarded program. goal %s in path [%s] is unguarded.\n", t.prettyString(gpidmap), trailPath(trail))
				os.Exit(1)
			}
		}
	}
	fmt.Println("Program is guarded.")
}

func subterms(t *Term, cl int) []context {
	c := make([]context, 0)
	p := make(path, 0)
	for i := range t.args {
		gatherSubterms(t.args[i], cl, append(p, i), &c)
	}
	return c
}

func gatherSubterms(t *Term, cl int, p path, c *[]context) {
	t = t.Deref()
	if !t.isVar() {
		np := make([]int, len(p))
		copy(np, p)
		*c = append(*c, context{cl, t, np})
		for i := range t.args {
			gatherSubterms(t.args[i], cl, append(p, i), c)
		}
	}
}

func recReducts(t1 *Term, t2 *Term, cl int) []context {
	c := make([]context, 0)
	p := make(path, 0)
	gatherRecursiveReducts(t1, t2, cl, p, &c)
	return c
}

func gatherRecursiveReducts(t1 *Term, t2 *Term, cl int, p path, c *[]context) {
	t1 = t1.Deref()
	t2 = t2.Deref()
	if t2.isVar() {
		if !t1.isVar() {
			tmap := make(TermMap)
			t1.CollectVids_(tmap)
			np := make([]int, len(p))
			copy(np, p)
			if _, ok := tmap[t2.Id()]; ok {
				*c = append(*c, context{cl, t1, np})
			}
		}
	} else {
		if t2.Id() == t1.Id() {
			for i := range t1.args {
				gatherRecursiveReducts(t1.args[i], t2.args[i], cl, append(p, i), c)
			}
		}
	}
}

func allReducts(t1 *Term, t2 *Term, cl int) []context {
	c := make([]context, 0)
	p := make(path, 0)
	gatherReducts(t1, t2, cl, p, &c)
	return c
}

func gatherReducts(t1 *Term, t2 *Term, cl int, p path, c *[]context) {
	t1 = t1.Deref()
	t2 = t2.Deref()
	if t2.isVar() || len(t2.args) == 0 {
		if !t1.isVar() && len(t1.args) > 0 {
			np := make([]int, len(p))
			copy(np, p)
			*c = append(*c, context{cl, t1, np})
		}
	} else {
		if t2.Id() == t1.Id() {
			for i := range t1.args {
				gatherReducts(t1.args[i], t2.args[i], cl, append(p, i), c)
			}
		}
	}
}

func isGuarded(tmap NodeTemplateMap, t *Term, goals []gtrail) (*Term, []gtrail) {
	debug := config.debug
	if debug {
		fmt.Printf("new goal: %s\n", t.prettyString(gpidmap))
	}

	tid := t.Id()

	// check unguarded loops
	if len(goals) > 0 {
		lastgoal := goals[len(goals)-1]

		if debug {
			fmt.Printf("searching for loops with predicate '%s' and clausenum %d to %d parent and-nodes...\n", gpidmap[tid], lastgoal.clause, len(goals)-1)
		}

		indices := make([]int, 0, len(goals)-1)
		for pos, goal := range goals[:len(goals)-1] {
			if pos > 0 {
				if tid == goal.id && goals[pos-1].clause == lastgoal.clause {
					indices = append(indices, pos)
				}
			}
		}

		if debug {
			if len(indices) > 0 {
				fmt.Printf("found total of %d loops from %d possible\n", len(indices), len(goals)-1)
			} else {
				fmt.Printf("found no loops\n")
			}
		}

		for _, pos := range indices {
			goal := goals[pos]
			if debug {
				fmt.Printf("found loop position %d to position %d\n", len(goals)-1, pos)
				fmt.Printf("parent %s ... child %s\n", goal.term.prettyString(gpidmap), t.prettyString(gpidmap))
			}
			// check if reduction happens
			reducts := allReducts(goal.term, t, goal.clause)
			if len(reducts) == 0 {
				if debug {
					fmt.Printf("is unguarded!\n")
				}
				return goal.term, goals
			}

			if debug {
				fmt.Printf("is guarded with %d recursive reducts: %s\n", len(reducts), reducts)
			}

		}
	}

clausesloop:
	for clausenum, nodetemplate := range tmap[tid] {

		if debug {
			fmt.Printf("testing clause %d for goal %s\n", clausenum+1, t.prettyString(gpidmap))
		}

		ntsmap := make(TermMap)
		ntsheadCopy := nodetemplate.Head.Copy_(ntsmap)

		if debug {
			fmt.Printf("std apart %s to %s with %s\n", nodetemplate.Head.prettyString(gpidmap), ntsheadCopy.prettyString(gpidmap), ntsmap)
			fmt.Printf("checking if %s is unifiable with %s\n", ntsheadCopy.prettyString(gpidmap), t.prettyString(gpidmap))
		}

		if ntsheadCopy.Unifiable(t) {

			if debug {
				fmt.Printf("%d path: [%s, (%s:%d)]\n", len(goals), trailPath(goals), gpidmap[tid], clausenum+1)
			}

			ttmap := make(TermMap)
			tCopy := t.Copy_(ttmap)

			if debug {
				fmt.Printf("computing subterms for %s\n", nodetemplate.Head.prettyString(gpidmap))
			}

			subterms := subterms(nodetemplate.Head, clausenum)

			if debug {
				fmt.Printf("found subterms: %s\n", subterms)
			}

			if debug {
				fmt.Printf("copy substitutions: %s\n", ttmap)
				fmt.Printf("goal before unification: %s with %s\n", tCopy.prettyString(gpidmap), ntsheadCopy.prettyString(gpidmap))
			}

			umap := make(TermMap)
			tCopy.UnifyMap_(ntsheadCopy, umap, config.occurscheck)

			if debug {
				fmt.Printf("umap: %s\n", umap)
				fmt.Printf("goal after unification: %s\n", tCopy.prettyString(gpidmap))
			}

			projections := make([]context, 0, len(subterms))

			if debug {
				fmt.Printf("computing reducts for %s and %s\n", tCopy.prettyString(gpidmap), t.prettyString(gpidmap))
			}

			reducts := allReducts(tCopy, t, clausenum)

			if debug {
				fmt.Printf("found reducts: %s\n", reducts)
			}

		outerreduct:
			for _, reduct := range reducts {
				for k, subterm := range subterms {
					if subterm.matches(reduct) {
						projections = append(projections, subterm)
						subterms = append(subterms[:k], subterms[k+1:]...)
						if len(subterms) == 0 {
							break outerreduct
						}
						break
					}
				}
			}

			if debug {
				fmt.Printf("found projections: %s\n", projections)
			}

			// copy goals
			goalsCopy := make([]gtrail, len(goals)+1)
			for i, g := range goals {
				goalsCopy[i] = gtrail{g.id, g.clause, g.term.Copy_(ttmap), g.projections, g.guardcontext}

			}
			goalsCopy[len(goalsCopy)-1] = gtrail{t.Id(), clausenum, tCopy, projections, make([]context, 0)}

			goalsold := goals
			goals := goalsCopy

			// find guarding context
			guardcontext := make([]context, 0)

			proj := make([]context, len(projections))
			copy(proj, projections)

			if len(goalsold) > 1 {
				lastgoal := goals[len(goals)-1]
				indices := make([]int, 0, len(goalsold))
				for pos, goalold := range goalsold {
					if tid == goalold.id && goalsold[pos].clause == lastgoal.clause {
						indices = append(indices, pos)
					}
				}

				if len(proj) > 0 {
				outer:
					for _, i := range indices {
						for _, j := range indices {
							if j >= i {
								break
							}
							g1 := goalsold[j]
							g2 := goalsold[i]
							if debug {
								fmt.Printf("computing reduction in loop %s to %s , %d to %d\n", g1.term.prettyString(gpidmap), g2.term.prettyString(gpidmap), i, j)
							}
							recreducts := recReducts(g1.term, g2.term, g1.clause)
							for _, reduct := range recreducts {
								if debug {
									fmt.Printf("checking reduction %s in loop %s to %s\n", reduct, g1.term.prettyString(gpidmap), g2.term.prettyString(gpidmap))
								}
								for k := range proj {
									//fmt.Printf("check projection %s\n", proj[k])
									if proj[k].matches(reduct) {
										if debug {
											fmt.Printf("found projection %s in recursive reduction %s for %s |> %s in loop from %d to %d\n", proj[k], reduct, g1.term.prettyString(gpidmap), g2.term.prettyString(gpidmap), j, i)
										}
										guardcontext = append(guardcontext, proj[k])
										proj = append(proj[:k], proj[k+1:]...)
										if len(proj) == 0 {
											break outer
										}
										break
									}
								}
							}
						}
					}
				}

				goals[len(goals)-1].guardcontext = guardcontext

				if len(guardcontext) > 0 {
					if debug {
						fmt.Printf("found guarding context: %s\n", guardcontext)
						fmt.Printf("checking if fixpoint...\n")
					}
					for _, i := range indices {
						if debug {
							fmt.Printf("comparing guarding contexts %s with %s\n", goals[i].guardcontext, guardcontext)
						}
						if equalContext(goals[i].guardcontext, guardcontext) {
							if debug {
								fmt.Printf("found guarding context fixpoint\n")
							}
							continue clausesloop
						}
					}

				} else {
					if debug {
						fmt.Printf("found no guarding context\n")
					}
				}

			}

			// recurse into body goals
			for _, ornode := range nodetemplate.Body.And {
				goalsNew := make([]gtrail, len(goals))
				copy(goalsNew, goals)
				if rterm, goals := isGuarded(tmap, ornode.Head.Copy_(ntsmap), goalsNew); rterm != nil {
					return rterm, goals
				}

			}
		}
	}
	return nil, nil
}
