package main

import (
	"fmt"
	"sync/atomic"
	"unicode"
)

const intSize = 32 << (^uint(0) >> 63)

const (
	typeBits = 1
	idBits   = (intSize - typeBits - 1)
)

const (
	typePos = 0
	idPos   = typeBits
)

const (
	tagVar = iota
	tagFunctor
)

const (
	typeMask = ((1 << typeBits) - 1) << typePos
	idMask   = ((1 << idBits) - 1) << idPos
)

type Cell int64

func (t *Term) Id() Id {
	return Id((t.cell & idMask) >> idPos)
}

func (t *Term) Type() int {
	return int((t.cell & typeMask) >> typePos)
}

func (t *Term) isVar() bool {
	return int((t.cell&typeMask)>>typePos) == tagVar
}

type PnameMap map[ptermKey]Id

func (pnames PnameMap) PidMap() PidMap {
	pidmap := make(PidMap)
	for key, value := range pnames {
		pidmap[value] = key.name
	}
	return pidmap
}

type PidMap map[Id]string

type PidCountMap map[Id]int

type VnameMap map[string]*Term

type Substitution struct {
	fromId Id
	to     *Term
}

type SubstList []Substitution

func (l SubstList) Len() int           { return len(l) }
func (l SubstList) Less(i, j int) bool { return l[i].fromId < l[j].fromId }
func (l SubstList) Swap(i, j int)      { l[i], l[j] = l[j], l[i] }

type TermMap map[Id]*Term

func (tmap TermMap) List() SubstList {
	list := make(SubstList, len(tmap))
	i := 0
	for id, to := range tmap {
		list[i] = Substitution{id, to}
		i++
	}
	return list
}

func (tmap TermMap) Copy() TermMap {
	tmapCopy := make(TermMap)
	for k, t := range tmap {
		tmapCopy[k] = t
	}
	return tmapCopy
}

func (tmap TermMap) Expand() TermMap {
	tmapNew := make(TermMap)
	for k, t := range tmap {
		tmap[k] = t.Copy_(tmapNew)
	}
	for k, t := range tmapNew {
		tmap[k] = t
	}
	return tmap
}

func (tmap TermMap) String() string {
	var s string
	if len(tmap) == 0 {
		return "{}"
	}
	for k, t := range tmap {
		s = fmt.Sprintf("%s,v%v -> %s", s, k, t.prettyString(gpidmap))
	}
	return "{" + s[1:] + "}"
}

func (tmap TermMap) Unlink() {
	for _, t := range tmap {
		t.link = nil
	}
}

type Term struct {
	cell Cell
	link *Term
	args []*Term
}

func (t *Term) Name() string {
	if t.isVar() {
		return fmt.Sprintf("v%v", t.Id())
	}
	return fmt.Sprintf("p%v", t.Id())
}

func (t *Term) String() string {
	s := fmt.Sprintf("%s", t.Name())
	if t.args != nil && len(t.args) != 0 {
		s = fmt.Sprintf("%s(%s", s, t.args[0])
		for _, a := range t.args[1:] {
			s = fmt.Sprintf("%s,%s", s, a)
		}
		s = s + ")"
	}
	return s
}

func (t *Term) prettyString(pids PidMap) string {
	t = t.Deref()
	if t.isVar() {
		return t.String()
	} else {
		name, ok := pids[t.Id()]
		if !ok {
			name = t.Name()
		}
		if len(t.args) == 0 {
			return name
		}
		s := fmt.Sprintf("%s(%s", name, t.args[0].prettyString(pids))
		for _, a := range t.args[1:] {
			s = fmt.Sprintf("%s,%s", s, a.prettyString(pids))
		}
		return s + ")"
	}

}

func (t *Term) Deref() *Term {
	for t.link != nil {
		t = t.link
	}
	return t
}

func (t *Term) Copy() *Term {
	return t.Copy_(make(TermMap))
}

func (t *Term) Copy_(tmap TermMap) *Term {
	t = t.Deref()
	id := t.Id()
	if t.isVar() {
		v, ok := tmap[id]
		if !ok {
			v = NewVariable()
			tmap[id] = v
		}
		return v
	} else {
		args := make([]*Term, len(t.args))
		for i, arg := range t.args {
			args[i] = arg.Copy_(tmap)
		}
		return NewPredicate(id, args)
	}
}

func (t *Term) CollectVids_(tmap TermMap) {
	t = t.Deref()
	if t.isVar() {
		tmap[t.Id()] = t
		return
	}
	for _, targ := range t.args {
		targ.CollectVids_(tmap)
	}
	return
}

func (t *Term) Equals(t2 *Term) bool {
	t1 := t.Deref()
	t2 = t2.Deref()

	if t1.Id() != t2.Id() {
		return false
	}

	if !t1.isVar() {
		for i := range t1.args {
			if !t1.args[i].Equals(t2.args[i]) {
				return false
			}
		}
	}
	return true
}

func (t *Term) Subsumes(specific *Term) bool {
	specificCopy := specific.Copy()
	specificCopy.NumberVars()
	return t.Unifiable(specificCopy)
}

func (t *Term) NumberVars() int {
	return -int(t.NumberVars_(Id(0)))
}

func (t *Term) NumberVars_(id Id) Id {
	t = t.Deref()
	if t.isVar() {
		id--
		t.link = NewPredicate(id, nil)
		return id
	} else {
		for _, targ := range t.args {
			id = targ.NumberVars_(id)
		}
	}
	return id
}

func (t *Term) Unifiable(t2 *Term) bool {
	tmap := make(TermMap)
	unifiable := t.UnifyMap_(t2, tmap, config.occurscheck)
	tmap.Unlink()
	return unifiable
}

func (t *Term) UnifiableNoCheck(t2 *Term) bool {
	tmap := make(TermMap)
	unifiable := t.UnifyMap_(t2, tmap, false)
	tmap.Unlink()
	return unifiable
}

func (t *Term) UnifyMap_(t2 *Term, tmap TermMap, occurscheck bool) bool {
	t1 := t.Deref()
	t2 = t2.Deref()

	id1 := t1.Id()
	id2 := t2.Id()

	if t1.isVar() {
		if t2.isVar() {
			if id1 < id2 {
				tmap[id2] = t2
				t2.link = t1
			} else if id1 > id2 {
				tmap[id1] = t1
				t1.link = t2
			}
		} else {
			if occurscheck && t2.Contains(id1) {
				return false
			}
			tmap[id1] = t1
			t1.link = t2
		}
	} else {
		if t2.isVar() {
			if occurscheck && t1.Contains(id2) {
				return false
			}
			tmap[id2] = t2
			t2.link = t1
		} else {
			if id1 != id2 {
				return false
			}
			for i := range t1.args {
				if !t1.args[i].UnifyMap_(t2.args[i], tmap, occurscheck) {
					return false
				}
			}

		}
	}
	return true
}

func (t *Term) Unify(t2 *Term) bool {
	tmap := make(TermMap)
	if !t.UnifyMap_(t2, tmap, config.occurscheck) {
		tmap.Unlink()
		return false
	}
	return true
}

func (t *Term) Match(t2 *Term) {
	t1 := t.Deref()
	t2 = t2.Deref()

	if t1.isVar() {
		if t2.isVar() {
			id1 := t1.Id()
			id2 := t2.Id()
			if id1 < id2 {
				t2.link = t1
			} else if id1 > id2 {
				t1.link = t2
			}
		} else {
			t1.link = t2
		}
	} else {
		for i := range t1.args {
			t1.args[i].Match(t2.args[i])
		}
	}
}

func (t *Term) CollectPids_(pidcountmap PidCountMap) {
	if !t.isVar() {
		pidcountmap[t.Id()] += 1
		for _, targ := range t.args {
			targ.CollectPids_(pidcountmap)
		}
	}
}

func (t *Term) isGround() bool {
	t = t.Deref()
	if t.isVar() {
		return false
	}
	for _, targ := range t.args {
		if !targ.isGround() {
			return false
		}
	}
	return true
}

func (t *Term) Contains(id Id) bool {
	if t.isVar() {
		if t.Id() == id {
			return true
		}
		if t.link != nil {
			return t.link.Contains(id)
		}
	} else {
		for _, targ := range t.args {
			if targ.Contains(id) {
				return true
			}
		}
	}
	return false
}

type ptermKey struct {
	name   string
	arrity int
}

func NewVariable() *Term {
	return &Term{NewVariableCell(Id(atomic.AddInt64((*int64)(&genid), 1))), nil, nil}
}

func NewVariableCell(id Id) Cell {
	return Cell(tagVar | (id << idPos))
}

func NewPredicate(id Id, args []*Term) *Term {
	return &Term{NewPredicateCell(id), nil, args}
}

func NewPredicateCell(id Id) Cell {
	return Cell(tagFunctor | (id << idPos))
}

func TermSubstitutions(t1 *Term, t2 *Term) TermMap {
	t1map := make(TermMap)
	t1.CollectVids_(t1map)
	t1.Unify(t2.Copy())
	substs := make(TermMap)
	for id, t := range t1map {
		if t.link != nil {
			substs[id] = t.link.Deref()
		}
	}
	t1map.Unlink()
	return substs
}

func isVariableName(s string) bool {
	return unicode.IsUpper(rune(s[0]))
}

func makeTerm(input string, pmap PnameMap) *Term {
	pterm, err := ParseTerm(input)
	if err != nil {
		panic(err)
	}
	return makeTermM(pterm, pmap, make(VnameMap))
}

func makeTermM(pterm *Pterm, pmap PnameMap, vmap VnameMap) *Term {
	if isVariableName(pterm.Name) {
		variable, ok := vmap[pterm.Name]
		if !ok {
			variable = NewVariable()
			vmap[pterm.Name] = variable
		}
		return variable
	}
	key := ptermKey{pterm.Name, len(pterm.Args)}
	id, ok := pmap[key]
	if !ok {
		id = Id(atomic.AddInt64((*int64)(&genid), 1))
		pmap[key] = id
	}
	t := NewPredicate(id, nil)
	if len(pterm.Args) > 0 {
		t.args = make([]*Term, len(pterm.Args))
		for i, ptermarg := range pterm.Args {
			t.args[i] = makeTermM(ptermarg, pmap, vmap)
		}
	}
	return t
}
